// $Id: lock.h,v 1.6 2006/06/20 17:50:45 dvermeir Exp $
#ifndef DV_THREAD_LOCK_H
#define DV_THREAD_LOCK_H

#include <string>
#include <iostream>
#include <dvthread/monitor.h>
#include <dvthread/thread.h>

namespace Dv {
  namespace Thread {
  
    /**
     * Convenience class to synchronize access to a Monitor object.
     * Creating a lock enters the monitor, destroying the lock exits the
     * monitor.
     *
     * Example usage:
     * @code
     * class SomeClassToProtect: public Dv::Thread::Monitor {
     * public:
     *   voif f() {
     *     Lock  lock(*this, "f", &cerr); // enter monitor
     *     access the object
     *     // destructor of lock will exit the monitor.
     *     }
     * ...
     * };
     * @endcode
     * 
     * The lock declaration can be abbreviated using the Java-inspired
     * @ref Dv::Thread::SYNCHRONIZED "SYNCHRONIZED" macro.
     */
    class Lock {
    public:
      /**
       * Constructor. This function will enter the monitor parameter.
       *
       * @param m monitor to enter.
       * @param msg optional message to print to log
       * @param log if nonzero, the message will be printed both
       *  on creation and destruction of the lock.
       * @see Monitor::enter
       */
      Lock(Monitor& m, const std::string& msg="", std::ostream* log=0);
      /**
       * Destructor. This function will exit the monitor.
       * @exception runtime_error if the current thread is not
       * in the monitor (e.g. if m.exit() was called before the
       * destructor of the Lock).
       * @see Monitor::exit()
       */
      ~Lock() throw (std::runtime_error);
    private:
      /** Monitor on which lock is defined.  */
      Monitor&  monitor_;
      /** Message to print on construction and destruction. */
      const std::string  msg_;
      /** If non-zero, pointer to stream on which to log info. */
      std::ostream*  log_;
    };
  }
}

/**
 * Convenience macro. Put this in the beginning of a member function
 * of a subclass of a Monitor object to ensure exclusive access to
 * some resource. Inspired by Java keyword.
 * @code
 * class Buffer: public Dv::Thread::Monitor {
 * public:
 *   int get() { SYNCHRONIZED retrieve & remove item; }
 *   void put(int) { SYNCHRONIZED add item; }
 * };
 * @endcode
 */
#define SYNCHRONIZED  Dv::Thread::Lock lock(*this);

#endif
