// $Id: monitor.C,v 1.5 2006/06/20 17:50:45 dvermeir Exp $

#include <stdlib.h>
#include <dvthread/monitor.h>
#include "ntostr.h"

#define TRACE std::cerr << pthread_self() << ":" << __FILE__ << "." << __LINE__ << std::endl;

static const std::string PREFIX("Dv::Thread::Monitor::");

void
Dv::Thread::Monitor::check(const std::string& msg) const throw (std::runtime_error) {
  static const std::string NAME(PREFIX + "check: ");
  if (! pthread_equal(owner_, pthread_self()) ) {
    throw std::runtime_error(NAME + msg + ": current thread #" +
      ntostr(pthread_self()) + " not owner (#" +
      ntostr(owner_) + ") of monitor " + name());
    }
}

void
Dv::Thread::Monitor::check_cond(const std::string& operation, unsigned int n) const throw (std::runtime_error) {
  static const std::string NAME(PREFIX + "check_cond");
  check(operation);
  if (n >= n_conditions_)
    throw std::runtime_error(NAME + name() + ": " + operation + " condition " + ntostr(n) 
       + ": bad condition, max is " + ntostr(n_conditions_ -1));
}

Dv::Thread::Monitor::Monitor(const std::string& nm, unsigned int n_conditions, std::ostream* dbg): 
  name_(nm), n_conditions_(n_conditions), conditions_(0), dbg_(dbg) {
  if (dbg_)
    *dbg_ << pthread_self() << ": " << PREFIX << "Monitor(" << name() 
          << ", " << n_conditions << ")" << std::endl;
  // create & initialize _mutex semaphore
  pthread_mutexattr_init(&mutex_attributes_); 
  pthread_mutex_init(&mutex_, &mutex_attributes_);
  pthread_condattr_init(&condition_attributes_); 
  // create & initialize conditions
  if (n_conditions_) {
    conditions_ = new pthread_cond_t[n_conditions_];
    for (unsigned int i=0;(i<n_conditions_);++i)
      pthread_cond_init(conditions_+i, &condition_attributes_);
  }
}

Dv::Thread::Monitor::~Monitor() {
  if (dbg_)
    *dbg_ << pthread_self() << ": " << PREFIX << "~Monitor()" << std::endl;
  pthread_mutex_destroy(&mutex_);
  for (unsigned int i=0;(i<n_conditions_);++i)
    pthread_cond_destroy(conditions_ + i);
  delete conditions_;
}

void
Dv::Thread::Monitor::enter() throw (std::runtime_error) {
  static const std::string NAME(PREFIX + "enter ");
  if (dbg_)
    *dbg_ << pthread_self() << ": " << "(" << NAME <<  name() << std::endl;
  if ( pthread_equal(owner_, pthread_self()) )
    throw std::runtime_error(NAME + name() + " by thread #"
     		        + ntostr(pthread_self()) + ": already owner");
  pthread_mutex_lock(&mutex_);
  owner_ = pthread_self();
}

void
Dv::Thread::Monitor::exit() throw (std::runtime_error) {
  static const std::string NAME(PREFIX + "exit ");
  if (dbg_)
    *dbg_ << pthread_self() << ": " << NAME <<  name() 
          << " by thread #" << pthread_self() << ")" << std::endl;
  check("exit");
  owner_ = 0;
  pthread_mutex_unlock(&mutex_);
}

void
Dv::Thread::Monitor::wait(unsigned int condition) throw (std::runtime_error) {
  static const std::string NAME(PREFIX + "wait on ");
  if (dbg_)
    *dbg_ << pthread_self() << ": " << "[" << NAME <<  name() << std::endl;
  // check_cond("wait", condition);
  pthread_cond_wait(conditions_ + condition, &mutex_);
  if (dbg_)
    *dbg_ << pthread_self() << ": " << NAME <<  name() << "." << ntostr(condition)
          << " by thread #" << pthread_self() << " done ]"<< std::endl;
  owner_ = pthread_self();
}

bool
Dv::Thread::Monitor::wait(unsigned int condition, time_t timeout) 
  throw (std::runtime_error) { 
  static const std::string NAME(PREFIX + "timed_wait on ");
  if (dbg_)
    *dbg_ << pthread_self() << ": " << "[" <<  NAME <<  name() << "." << condition << std::endl;
  check_cond("timed_wait", condition);
  struct timeval now;
  gettimeofday(&now,0);
  timespec t_spec;
  t_spec.tv_sec = now.tv_sec + timeout/1000; // # of secs
  // 1 nsec = 10^-9 sec (nanosec).
  t_spec.tv_nsec = now.tv_usec * 1000 + (timeout % 1000) * 1000000; 
  int n = pthread_cond_timedwait(conditions_ + condition, &mutex_, &t_spec);
  if (dbg_)
    *dbg_ << "]" <<  pthread_self() << ": " << NAME <<  name() << "." << condition << ": " << n << std::endl;
  owner_ = pthread_self();
  return (n == 0);
}

void
Dv::Thread::Monitor::signal(unsigned int condition) const throw (std::runtime_error) {
  static const std::string NAME(PREFIX + "signal on ");
  if (dbg_)
    *dbg_ << pthread_self() << ": " << NAME <<  name() << ", condition " << condition << std::endl;
  check_cond("signal", condition);
  pthread_cond_signal(conditions_ + condition);
}

void
Dv::Thread::Monitor::broadcast(unsigned int condition) const throw (std::runtime_error) {
  static const std::string NAME(PREFIX + "broadcast on ");
  if (dbg_)
    *dbg_ << pthread_self() << ": " << NAME <<  name() << ", condition " << condition << std::endl;
  check_cond("signal", condition);
  pthread_cond_broadcast(conditions_ + condition);
}

