#ifndef DV_THREAD_MONITOR_H
#define DV_THREAD_MONITOR_H
// $Id: monitor.h,v 1.11 2006/06/20 17:50:45 dvermeir Exp $

#include <string>
#include <iostream>
#include <stdexcept>
extern "C" {
#include <pthread.h> // posix thread library
}
#include <sys/time.h>

namespace Dv {
  namespace Thread {
    /**
     * A simple Monitor class using posix threads.
     *
     * The following defines a Buffer monitor with protected
     * get() and put() access.
     * @include buffer.h
     * 
     * @see Dv::Thread::Thread
     */
    class Monitor {
    public:
      /** 
       * Constructor.
       * @param name of this monitor.
       * @param n_conditions number of conditions in this monitor.
       * @param dbg if non-zero, debug output will be written to @a *dbg.
       */
      Monitor(const std::string& name, unsigned int n_conditions=0, std::ostream* dbg = 0);
      /**
       * Virtual destructor.
       */
      virtual ~Monitor();
      /**
       * Enter the monitor. 
       * @exception runtime_error if this thread already owns the monitor
       */
      void enter() throw (std::runtime_error);
      /**
       * Exit the monitor.
       * @exception runtime_error if this thread is not the current owner
       *   of the monitor
       */
      void exit() throw (std::runtime_error);
      /**
       * Unconditionally wait for a given condition.
       * @param condition number of the condition to wait for.
       * @exception runtime_error if this thread is not the current owner
       *   of the monitor
       */
      void wait(unsigned int condition=0) throw (std::runtime_error); 
      /**
       * Wait for a given condition or until timeout millisecs elapsed.
       *
       * @param condition number of the condition to wait for.
       * @param timeout number of milliseconds we are prepared to wait.
       * @return false if a timeout occured.
       * @exception runtime_error if this thread is not the current owner
       *   of the monitor
       */
      bool wait(unsigned int condition, time_t timeout) throw (std::runtime_error); 
      /**
       * Signal a given condition. This function unblocks at least one
       * thread blocked on the condition. The unblocked threads contend
       * for the monitor.
       *
       * @param condition to signal.
       * @exception runtime_error if this thread is not the current owner
       *   of the monitor
       */
      void signal(unsigned int condition=0) const throw (std::runtime_error);
      /**
       * Signal all threads waiting on this condition.
       * This function unblocks all the threads blocked on the condition.
       * The unblocked threads contend for the monitor.
       *
       * @param condition to signal.
       * @exception runtime_error if this thread is not the current owner
       *   of the monitor
       */
      void broadcast(unsigned int condition=0) const throw (std::runtime_error); 
      /**
       * @return name of this monitor.
       */
      const std::string& name() const { return name_; }
    private:
      /**
       * Name of the monitor.
       */
      std::string name_; 
      /**
       * Thread that currently owns the monitor.
       */
      pthread_t owner_; 
      /**
       * Entry to the monitor is controlled by this mutex semaphore.
       */
      pthread_mutex_t mutex_;
      /**
       * Attributes of mutex_.
       */
      pthread_mutexattr_t mutex_attributes_;
      /**
       * Number of conditions of the monitor.
       */
      unsigned int n_conditions_; 
      /**
       * Array of n_conditions_ posix conditions of this monitor.
       */
      pthread_cond_t*  conditions_;
      pthread_condattr_t  condition_attributes_;
      /**
       * This functions throws a runtime_error if the current thread
       * is not the current owner of the monitor.
       *
       * @param message used as part of runtime_error.what()
       * @exception runtime_error if this thread is not the current owner
       *   of the monitor
       */
      void check(const std::string& message) const throw (std::runtime_error); 
      /**
       * This functions throws a runtime_error if n is not a valid
       * condition of this monitor or if the current thread does not
       * own the monitor.
       * @param msg used as part of runtime_error.what()
       * @param n number of condition to check
       * @exception runtime_error if this thread is not the current owner
       *   of the monitor
       */
      void check_cond(const std::string& msg,unsigned int n) const 
        throw (std::runtime_error); 
      /**
       * No copy ctor.
       */
      Monitor(const Monitor&); 
      /**
       * No assignment operator.
       */
      Monitor& operator=(const Monitor&); 
      /** Write debug output to stream if not 0 */
      std::ostream* dbg_;
    };
  }
}
  
#endif
