// $Id: thread.C,v 1.11 2006/06/24 09:39:28 dvermeir Exp $

#include <map>
#include <iostream>
#include <errno.h>
#include <signal.h>
#include <unistd.h> // usleep

#include <dvthread/lock.h>
#include <dvthread/thread.h>
#include "ntostr.h"

#define TRACE std::cerr << pthread_self() << ": " << __FILE__ << "." << __LINE__ << std::endl;

std::ostream* Dv::Thread::Thread::log = 0; 

static std::ostream&
LOG() {
  if (Dv::Thread::Thread::log) {
    *Dv::Thread::Thread::log << "pthread " << pthread_self() << ": ";
    Dv::Thread::Thread::log->flush();
    return *Dv::Thread::Thread::log;
  }
  else
    return std::cerr;
}

int
Dv::Thread::Thread::main() {
  if (Dv::Thread::Thread::log)
    LOG() << "running default main" << std::endl;
  return DEFAULT_MAIN;
}

void
Dv::Thread::Thread::set_signal_handler_installed() {
  if (is_self())
    signal_handler_installed = true;
}

///////////////// Signal handler for SIGFPE
void dv_thread_handle_sigterm(int) {
  signal(SIGFPE, dv_thread_handle_sigterm);
  if (Dv::Thread::Thread::log)
    LOG() << "caught SIGFPE" << std::endl;
  Dv::Thread::Thread::self()->killed();
  if (Dv::Thread::Thread::log)
    LOG() << "handled SIGFPE" << std::endl;
}

///////////////// Auxiliary function to start thread.
extern "C" void* dv_thread_run_thread(Dv::Thread::Thread* t) {
  static const std::string NAME("dv_thread_run_thread");
  // wait for id() to have been set by parent thread
  // wait for id() to have been set by parent thread
  while (t->id() == 0)
    usleep(1000); // sleep for 1 millisec
  while (t->id() == 0)
    usleep(1000); // sleep for 1 millisec
  // Install signal handler.
  signal(SIGFPE, dv_thread_handle_sigterm);
  t->set_signal_handler_installed();
  try {
    t->set_status(t->main());
  }
  catch (int status) {
    t->set_status(status);
  }
  catch (std::exception& e) {
    t->set_status(Dv::Thread::Thread::BAD_EXCEPTION);
    if (Dv::Thread::Thread::log)
      LOG() << NAME << ": bad exception: " << e.what() << std::endl;
  }
  catch (...) {
    t->set_status(Dv::Thread::Thread::BAD_EXCEPTION);
    if (Dv::Thread::Thread::log)
      LOG() << NAME << ": unknown exception" << std::endl;
  }
  if (t->delete_after_main())
    delete t;
  return 0;
}


///////////////// Static map<pthread, Thread>, access protected by Monitor
static std::map<pthread_t, Dv::Thread::Thread*>&
threads() {
  static std::map<pthread_t,Dv::Thread::Thread*> threads_; // currently known threads
  return threads_;
}

///////////////// Static Monitor used to protect access to threads map.
static Dv::Thread::Monitor&
monitor() {
  // protects access to threads() std::map
  static Dv::Thread::Monitor monitor_("Dv::Thread::threads"); 
  return monitor_;
}

///////////////// Constructor.
Dv::Thread::Thread::Thread(bool delete_after_main) throw (std::runtime_error): 
    id_(0), exit_status_(0), delete_after_main_(delete_after_main), signal_handler_installed(false) {
  if (log)
    LOG() << "Thread::Thread(delete_after_main=" << (delete_after_main ? "true": "false") << ")" << std::endl;
}

std::string
Dv::Thread::Thread::sid() const {
  return ntostr(id());
}

///////////////// Detach this thread
bool
Dv::Thread::Thread::is_self() const {
  pthread_t running_thread = pthread_self();
  if (running())
    return pthread_equal(running_thread, id_);
  return false;
}

///////////////// Detach this thread
Dv::Thread::Thread&
Dv::Thread::Thread::detach() {
  pthread_detach(id());
  return *this;
}

///////////////// Launch this tread.
int
Dv::Thread::Thread::start() throw (std::runtime_error) {
  static const std::string NAME("Thread::start");

  if (running()) 
    throw std::runtime_error(NAME + ": " + sid() + " already running");
  
  typedef void * (*VOID_FUNC_PTR)(void *);

  VOID_FUNC_PTR start_routine = reinterpret_cast<VOID_FUNC_PTR>(dv_thread_run_thread);

  if (pthread_create( &id_, 0, start_routine, this)) {
    int err(errno);
    std::string error_msg(strerror(err));
    throw std::runtime_error(NAME + ":pthread_create: " + error_msg + " (" + ntostr(err) + ")");
    }

  if (log)
    LOG() << NAME << "(): id = " << id() << std::endl;
  
  // start critical section
  Dv::Thread::Lock lock(monitor()); 
  threads()[id_] = this;
  return 0;
}

///////////////// Wait for this thread to finish.
int
Dv::Thread::Thread::join() const throw () {
  static const std::string NAME("Thread::join");
  // from man pthread_join:
  //
  // ERRORS
  //   ESRCH  No  thread could be found corresponding to that specified by th.
  //   EINVAL The th thread has been detached.
  //   EINVAL Another thread is already waiting on termination of th.
  //   EDEADLK The th argument refers to the calling thread.
  if (log)
    LOG() << NAME << "[ (id=" << id_ << ")" << std::endl;
  int status = pthread_join(id_, NULL);
  if (log)
    LOG() << NAME << "(id=" << id_ << "): " << status << "]" << std::endl;
  return status;
}

///////////////// Signal (SIGFPE) this thread.
int
Dv::Thread::Thread::kill() const throw (std::runtime_error) {
  static const std::string NAME("Thread::kill: ");
  if (log)
    LOG() << NAME << "(id=" << id_ << ")" << std::endl;
  // wait for id() to have been set by parent thread
  while (! signal_handler_installed)
    usleep(1000); // sleep for 1 millisec
  if (!running()) 
    throw std::runtime_error(NAME + sid() + " not running");
  if (is_self()) 
    throw std::runtime_error(NAME + "suicide not allowed");
  return pthread_kill(id_, SIGFPE);
}

///////////////// Return pointer to currently running thread.
Dv::Thread::Thread*
Dv::Thread::Thread::self() throw (std::runtime_error) {
  // Return pointer to current thread, may be 0 if thread made somewhere else
  pthread_t t = pthread_self();
  Lock lock(monitor()); // start critical section
  std::map<pthread_t,Thread*>::const_iterator i = threads().find(t);
  if (i==threads().end()) 
    return 0;
  return (*i).second;
}

///////////////// Finish thread with exit status.
void
Dv::Thread::Thread::exit(int status) {
  static const std::string NAME("Thread::exit");
  if (log)
    LOG() << NAME << "(status=" << status << ")" << std::endl;
  throw status;
}

///////////////// Destructor
Dv::Thread::Thread::~Thread() {
  static const std::string NAME("Thread::~Thread");
  if (log)
    LOG() << NAME << "(id=" << id_ << ", this = " << this << ")" << std::endl;
  // Join the thread, just in case. This should have been done in the destructor of the subclass
  // The call may fail e.g. if detached or if called from itself (e.g. by
  // dv_thread_run_thread, but who cares?
  // pthread_join(id_, NULL); 
  join();
  Lock lock(monitor()); 
  if (id_)
    threads().erase(id_); 
  if (log)
    LOG() << NAME << "(id=" << id_ << ") DONE" << std::endl;
}


