// $Id: thread.h,v 1.15 2006/06/24 09:39:28 dvermeir Exp $
#ifndef DV_THREAD_THREAD_H
#define DV_THREAD_THREAD_H

#include <setjmp.h>
#include <dvthread/monitor.h>
#include <iostream>

/**
 * @mainpage
 *
 * @author dvermeir@vub.ac.be
 *
 * <a href="../download/">Download</a>
 *
 * @section Installation Installation
 *
 * Standard. See the INSTALL file in the top directory.
 *
 * @section Description Description
 * This package provides two main classes: Dv::Thread::Thread,
 * representing threads, and Dv::Thread::Monitor, to represent
 * monitors. In addition, there is a trivial convenience class
 * Dv::Thread::Lock that can be used to conveniently ensure
 * exclusive access for member functions to *this (where *this
 * is an instance of (a subclass of) Dv::Thread::Monitor.
 *
 * @section Example Example program
 *
 * The following class implements a buffer for concurrent access.
 * Note the use of the SYNCHRONIZED macro which ensures that the get()
 * and put() member functions have exclusive access to the Buffer
 * object. Conditions are used to signal threads waiting to put or get
 * items into (from) the buffer.
 * @include buffer.h 
 *
 * The example below shows the implementation of a reader thread class
 * for the above buffer.
 * Each thread should implement a Thread::main() member function
 * that describes the work to be done by the thread.
 * @include reader.h
 *
 * This example shows the implementation of a writer thread class.
 * Each thread should implement a Thread::main() member function
 * that describes the work to be done by the thread.
 * @include writer.h
 *
 * The main program creates a Buffer object, a Reader and
 * a Writer thread. Note that the writer wants to put more items
 * than will be read. Consequently, it will throw an exception (from
 * within Buffer::put()).
 * @include test-thread.C
 */
namespace Dv {
  namespace Thread {
    /**
     * Posix thread class.
     *
     * Example Thread class definitions:
     * <ul>
     * <li> A concurrent Buffer reader class.
     * \include reader.h
     * <li> A concurrent Buffer writer class.
     * \include writer.h
     * </ul>
     * \see Dv::Thread::Monitor
     */
    class Thread {
    public:
      enum errors {
        BAD_EXCEPTION = -1 /** status of thread that threw a non-int exception */,
        DEFAULT_MAIN = -2 /** status of thread run with default main() */
      };

      /** Constructor.
       * @param delete_after_main if true, \a delete(t) will
       *   be executed after \a t->main() has completed.
       * @sa Dv::Thread::Thread::delete_after_main
       */
      Thread(bool delete_after_main = false) throw (std::runtime_error);
    
      /** Destructor. */
      virtual ~Thread();
      /** Start up thread. This function will execute main().
       * @return 0 iff ok, system errno else.
       * @exception runtime_error if thread is already running.
       */
      int start() throw (std::runtime_error); 

      /** Main function of thread.
       * @return the value to keep in set_status
       * Note that this function should not throw any exceptions
       * except integers.
       * The library will catch any exception e that is thrown erroneously,
       * print e.what() on cerr and return an exit status of BAD_EXCEPTION.
       * @sa Dv::Thread::Thread::start
       * @sa Dv::Thread::Thread::BAD_EXCEPTION
       * @sa Dv::Thread::Thread::exit
       * @sa Dv::Thread::Thread::set_status
       */
      virtual int main();
    
      /** This function is called by the thread as a response to a SIGFPE signal being
       * delivered from another thread, using Thread::kill. This function
       * is then called  the internal handler for this signal.
       * @warning This function should not throw any exceptions or chaos will
       * result. Typical usage might be:
       * @code
       * class MyThread {
       * public:
       *  ...
       *  void main() throw() {
       *    while (!killed_ && socket_>>line) {
       *      ...
       *      }
       *    }
       *  void killed() throw () {
       *    killed_ = true;
       *    socket_.close();
       *    }
       * }
       * @endcode
       */
      virtual void killed() throw () { return; }
      /** Wait for this thread to finish.
       * @return 0 iff ok, see manpage for @a pthread_self for other
       *   possibities.
       * @warning only works for non-detached threads.
       * @sa Dv::Thread::Thread::detach
       * @code
       * MyThread t;
       * cout << "starting t" << endl;
       * t.start();
       * cout << "waiting for t to finish" << endl;
       * t.join();
       * @endcode
       */
      int join() const throw (); 
    
      /** Kill a thread. This function delivers a <code>SIGFPE</code> signal to the
       * thread using <code>pthread_kill(id_,SIGFPE1)</code>. The signal is caught
       * and handled internally by calling Dv::Thread::Thread::killed().
       * @exception runtime_error for a thread that tries to kill itself.
       * @sa Dv::Thread::Thread::killed
       */
      int kill() const throw (std::runtime_error); 
      /** Return exit status of this thread.
       * @return exit status of thread.
       */
      int status() const { return exit_status_; }
      void set_status(int status) { exit_status_ = status; }
    
      /** Return posix thread id of this thread.
       * @return Posix thread id.
       */
      pthread_t id() const { return id_; }

      /** String representation of thread id of this thread.
       * @return string representation of the id of this thread
       */
      std::string sid() const;
    
      /** Detach this thread. 
       * Threads are created as ``joinable'', this function makes this
       * impossible. It also ensures that all resources are freed when the
       * thread exits.
       * @warning To avoid a memory leak, a thread must be joined or
       *   detached.
       * @return *this
       * @sa Dv::Thread::Thread::join
       */
      Thread& detach();
    
      /** Delete this thread object after Thread::main() has completed.
       * This feature is handy when there are an unknown number of threads
       * that are created using new and the only thing to do after a
       * thread has finished is to delete the Thread object.
       * 
       * The code below shows a typical example.
       * @code
       * class MyThread: public Dv::Thread::Thread {
       * public:
       *   MyThread* make(); // factory method to prohibit non-heap allocation.
       *   void main();
       * private:
       *   MyThread(...): Dv::Thread::Thread(true), .. { detach(); .. }  
       * };
       *
       * while (accept) {
       *   MyThread::make()->start(); // will delete itself automatically
       *   }
       * @endcode
       * @param yes if true, the object will be deleted after Thread::main
       *   has finished.
       * @return *this
       */
      Thread& delete_after_main(bool yes) { delete_after_main_ = yes; return *this; }
    
      /** Will this thread object be deleted after Thread::main() has completed?
       * @return true iff this object will be deleted after Thread::main()
       *   finished.
       */
      bool delete_after_main() { return delete_after_main_; }
    
      /** Is this thread the currently running thread?
       * @return true iff this thread is the currently running thread
       * @return false otherwise
       */
      bool is_self() const;

      /** Is this thread running?
       * @return true iff this thread has been created using pthread_create
       * @return false otherwise
       */
      bool running() const { return id_ != 0; }

      /** Return pointer to currently running thread.
       * @return pointer to currently running thread, may be 0
       *   if the thread was not made by this library.
       */
      static Thread* self() throw (std::runtime_error);
    
      /** Return id (pthread_t) of currently running thread.
       * This works for any thread; it is actually just call to
       * pthread_self.
       * @return id of currently running thread.
       */
      static pthread_t self_id() throw () { return pthread_self(); }
    
      /** Finish this thread, with status.
       * @warning This is implemented by throwing the status integer
       * as an exception. Hence is your code catches integers,
       * as in
       * @code
       * try {
       *    exit(3);
       * }
       * catch (integer e) {
       * }
       * @endcode
       * the thread will not exit.
       * 
       * The implementation of this function was changed between
       * versions 0.4 and 0.5 because the previous implementation,
       * which used @a pthread_exit, gave problems on systems with
       * the nptl thread implementation.
       * @param status to return.
       */
      static void exit(int status); 
    
      /** Stream to log info on, if not 0. */
      static std::ostream* log;
    
    protected:
      /** This function is called by the thread as a response to a SIGFPE signal being
       * delivered from another thread, using Dv::Thread::Thread::signal. The signal handler
       * is internal.
       * Typical usage might be:
       * @code
       * class MyThread {
       * public:
       *  ...
       *  void main() throw() {
       *    while (!killed_ && socket_>>line) {
       *      ...
       *      }
       *  void killed() {
       *    killed_ = true;
       *    socket_.close();
       *    }
       * }
       * @endcode
       * @warning This function should not throw any exceptions or chaos will
       * result, because exception throwing from signal handlers is
       * currently not supported for some cpu types.
       */
    private:
      /** Posix thread id. */
      pthread_t  id_; 
      /** Posix thread attributes.  */
      pthread_attr_t  attributes_;
      /** Exit status of this thread.  */
      int  exit_status_; 
      /** Forbid cctor. */
      Thread(const Thread&); 
      /** Forbid assignment.  */
      Thread& operator=(const Thread&); 
      /** Should this object be deleted after Thread::main has completed?  */
      bool delete_after_main_;
      /** Is a signal handler for SIGFPE installed? This should really
       * be private but it is not since it is set from a C function */
      bool signal_handler_installed;
    public:
      /** Set signal_handler_installed. This function does nothing if
       * not called by the thread itself.
       * @sa Dv::Thread::Thread::is_self
       */
      void set_signal_handler_installed();
    };
  }
}
#endif
